@@ -22,13 +22,13 @@ import sys
 import time
 import uuid
 
-from positional import positional
 import requests
 import six
 from six.moves import urllib
 
 import keystoneauth1
 from keystoneauth1 import _utils as utils
+from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 
 try:
@@ -45,7 +45,10 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
     keystoneauth1.__version__, requests.utils.default_user_agent(),
     platform.python_implementation(), platform.python_version())
 
-_logger = utils.get_logger(__name__)
+# NOTE(jamielennox): Clients will likely want to print more than json. Please
+# propose a patch if you have a content type you think is reasonable to print
+# here and we'll add it to the list as required.
+_LOG_CONTENT_TYPES = set(['application/json'])
 
 
 def _construct_session(session_obj=None):
@@ -60,6 +63,42 @@ def _construct_session(session_obj=None):
     return session_obj
 
 
+def _mv_legacy_headers_for_service(mv_service_type):
+    """Workaround for services that predate standardization.
+
+    TODO(sdague): eventually convert this to using os-service-types
+    and put the logic there. However, right now this is so little
+    logic, inlining it for release is a better call.
+
+    """
+    headers = []
+    if mv_service_type == "compute":
+        headers.append("X-OpenStack-Nova-API-Version")
+    elif mv_service_type == "baremetal":
+        headers.append("X-OpenStack-Ironic-API-Version")
+    return headers
+
+
+def _sanitize_headers(headers):
+    """Ensure headers are strings and not bytes."""
+    str_dict = {}
+    for k, v in headers.items():
+        if six.PY3:
+            # requests expects headers to be str type in python3, which means
+            # if we get a bytes we need to decode it into a str
+            k = k.decode('ASCII') if isinstance(k, six.binary_type) else k
+            if v is not None:
+                v = v.decode('ASCII') if isinstance(v, six.binary_type) else v
+        else:
+            # requests expects headers to be str type in python2, which means
+            # if we get a unicode we need to encode it to ASCII into a str
+            k = k.encode('ASCII') if isinstance(k, six.text_type) else k
+            if v is not None:
+                v = v.encode('ASCII') if isinstance(v, six.text_type) else v
+        str_dict[k] = v
+    return str_dict
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -134,11 +173,11 @@ def _determine_calling_package():
             # hit the bottom of the frame stack
             break
 
-    return None
+    return ''
 
 
 def _determine_user_agent():
-    """Attempt to programatically generate a user agent string.
+    """Attempt to programmatically generate a user agent string.
 
     First, look at the name of the process. Return this unless it is in
     the `ignored` list.  Otherwise, look at the function call stack and
@@ -153,10 +192,10 @@ def _determine_user_agent():
         name = sys.argv[0]
     except IndexError:
         # sys.argv is empty, usually the Python interpreter prevents this.
-        return None
+        return ''
 
     if not name:
-        return None
+        return ''
 
     name = os.path.basename(name)
     if name in ignored:
@@ -164,6 +203,24 @@ def _determine_user_agent():
     return name
 
 
+class RequestTiming(object):
+    """Contains timing information for an HTTP interaction."""
+
+    #: HTTP method used for the call (GET, POST, etc)
+    method = None
+
+    #: URL against which the call was made
+    url = None
+
+    #: Elapsed time information
+    elapsed = None  # type: datetime.timedelta
+
+    def __init__(self, method, url, elapsed):
+        self.method = method
+        self.url = url
+        self.elapsed = elapsed
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
@@ -207,6 +264,27 @@ class Session(object):
                                     to every request passing through the
                                     session. Headers of the same name specified
                                     per request will take priority.
+    :param str app_name: The name of the application that is creating the
+                         session. This will be used to create the user_agent.
+    :param str app_version: The version of the application creating the
+                            session. This will be used to create the
+                            user_agent.
+    :param list additional_user_agent: A list of tuple of name, version that
+                                       will be added to the user agent. This
+                                       can be used by libraries that are part
+                                       of the communication process.
+    :param dict discovery_cache: A dict to be used for caching of discovery
+                                 information. This is normally managed
+                                 transparently, but if the user wants to
+                                 share a single cache across multiple sessions
+                                 that do not share an auth plugin, it can
+                                 be provided here. (optional, defaults to
+                                 None which means automatically manage)
+    :param bool split_loggers: Split the logging of requests across multiple
+                               loggers instead of just one. Defaults to False.
+    :param bool collect_timing: Whether or not to collect per-method timing
+                                information for each API call. (optional,
+                                defaults to False)
     """
 
     user_agent = None
@@ -215,10 +293,12 @@ class Session(object):
 
     _DEFAULT_REDIRECT_LIMIT = 30
 
-    @positional(2)
     def __init__(self, auth=None, session=None, original_ip=None, verify=True,
                  cert=None, timeout=None, user_agent=None,
-                 redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None):
+                 redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
+                 app_name=None, app_version=None, additional_user_agent=None,
+                 discovery_cache=None, split_loggers=None,
+                 collect_timing=False):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -228,19 +308,22 @@ class Session(object):
         self.timeout = None
         self.redirect = redirect
         self.additional_headers = additional_headers or {}
+        self.app_name = app_name
+        self.app_version = app_version
+        self.additional_user_agent = additional_user_agent or []
+        self._determined_user_agent = None
+        if discovery_cache is None:
+            discovery_cache = {}
+        self._discovery_cache = discovery_cache
+        # NOTE(mordred) split_loggers kwarg default is None rather than False
+        # so we can distinguish between the value being set or not.
+        self._split_loggers = split_loggers
+        self._collect_timing = collect_timing
+        self._api_times = []
 
         if timeout is not None:
             self.timeout = float(timeout)
 
-        # Per RFC 7231 Section 5.5.3, identifiers in a user-agent should be
-        # ordered by decreasing significance.  If a user sets their product
-        # that value will be used. Otherwise we attempt to derive a useful
-        # product value. The value will be prepended it to the KSA version,
-        # requests version, and then the Python version.
-
-        if user_agent is None:
-            user_agent = _determine_user_agent()
-
         if user_agent is not None:
             self.user_agent = "%s %s" % (user_agent, DEFAULT_USER_AGENT)
 
@@ -280,25 +363,52 @@ class Session(object):
     def _process_header(header):
         """Redact the secure headers to be logged."""
         secure_headers = ('authorization', 'x-auth-token',
-                          'x-subject-token',)
+                          'x-subject-token', 'x-service-token')
         if header[0].lower() in secure_headers:
-            token_hasher = hashlib.sha1()
+            token_hasher = hashlib.sha1()  # nosec log hashing
             token_hasher.update(header[1].encode('utf-8'))
             token_hash = token_hasher.hexdigest()
             return (header[0], '{SHA1}%s' % token_hash)
         return header
 
-    @positional()
+    def _get_split_loggers(self, split_loggers):
+        """Get a boolean value from the various argument sources.
+
+        We default split_loggers to None in the kwargs of the Session
+        constructor so we can track set vs. not set. We also accept
+        split_loggers as a parameter in a few other places. In each place
+        we want the parameter, if given by the user, to win.
+        """
+        # None is the default value in each method's kwarg. None means "unset".
+        if split_loggers is None:
+            # If no value was given, try the value set on the instance.
+            split_loggers = self._split_loggers
+        if split_loggers is None:
+            # If neither a value was given on the method, nor a value was given
+            # on the Session constructor, then the value is False.
+            split_loggers = False
+        return split_loggers
+
     def _http_log_request(self, url, method=None, data=None,
                           json=None, headers=None, query_params=None,
-                          logger=_logger):
+                          logger=None, split_loggers=None):
+        string_parts = []
+
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.request')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('REQ:')
+            if not logger:
+                logger = utils.get_logger(__name__)
+
         if not logger.isEnabledFor(logging.DEBUG):
             # NOTE(morganfainberg): This whole debug section is expensive,
             # there is no need to do the work if we're not going to emit a
             # debug log.
             return
 
-        string_parts = ['REQ: curl -g -i']
+        string_parts.append('curl -g -i')
 
         # NOTE(jamielennox): None means let requests do its default validation
         # so we need to actually check that this is False.
@@ -321,7 +431,8 @@ class Session(object):
             string_parts.append(url)
 
         if headers:
-            for header in six.iteritems(headers):
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
                 string_parts.append('-H "%s: %s"'
                                     % self._process_header(header))
         if json:
@@ -336,10 +447,20 @@ class Session(object):
 
         logger.debug(' '.join(string_parts))
 
-    @positional()
     def _http_log_response(self, response=None, json=None,
                            status_code=None, headers=None, text=None,
-                           logger=_logger):
+                           logger=None, split_loggers=True):
+        string_parts = []
+        body_parts = []
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.response')
+            body_logger = utils.get_logger(__name__ + '.body')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('RESP:')
+            body_parts.append('RESP BODY:')
+            body_logger = logger
+
         if not logger.isEnabledFor(logging.DEBUG):
             return
 
@@ -348,30 +469,99 @@ class Session(object):
                 status_code = response.status_code
             if not headers:
                 headers = response.headers
-            if not text:
-                text = self._remove_service_catalog(response.text)
-        if json:
-            text = self._json.encode(json)
-
-        string_parts = ['RESP:']
 
         if status_code:
             string_parts.append('[%s]' % status_code)
         if headers:
-            for header in six.iteritems(headers):
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
                 string_parts.append('%s: %s' % self._process_header(header))
-        if text:
-            string_parts.append('\nRESP BODY: %s\n' % text)
-
         logger.debug(' '.join(string_parts))
 
-    @positional()
+        if not body_logger.isEnabledFor(logging.DEBUG):
+            return
+
+        if response is not None:
+            if not text:
+                # NOTE(samueldmq): If the response does not provide enough info
+                # about the content type to decide whether it is useful and
+                # safe to log it or not, just do not log the body. Trying to
+                # read the response body anyways may result on reading a long
+                # stream of bytes and getting an unexpected MemoryError. See
+                # bug 1616105 for further details.
+                content_type = response.headers.get('content-type', None)
+
+                # NOTE(lamt): Per [1], the Content-Type header can be of the
+                # form Content-Type := type "/" subtype *[";" parameter]
+                # [1] https://www.w3.org/Protocols/rfc1341/4_Content-Type.html
+                for log_type in _LOG_CONTENT_TYPES:
+                    if content_type is not None and content_type.startswith(
+                            log_type):
+                        text = self._remove_service_catalog(response.text)
+                        break
+                else:
+                    text = ('Omitted, Content-Type is set to %s. Only '
+                            '%s responses have their bodies logged.')
+                    text = text % (content_type, ', '.join(_LOG_CONTENT_TYPES))
+        if json:
+            text = self._json.encode(json)
+
+        if text:
+            body_parts.append(text)
+            body_logger.debug(' '.join(body_parts))
+
+    @staticmethod
+    def _set_microversion_headers(
+            headers, microversion, service_type, endpoint_filter):
+        # We're converting it to normalized version number for two reasons.
+        # First, to validate it's a real version number. Second, so that in
+        # the future we can pre-validate that it is within the range of
+        # available microversions before we send the request.
+        # TODO(mordred) Validate when we get the response back that
+        # the server executed in the microversion we expected.
+        # TODO(mordred) Validate that the requested microversion works
+        # with the microversion range we found in discovery.
+        microversion = discover.normalize_version_number(microversion)
+        # Can't specify a M.latest microversion
+        if (microversion[0] != discover.LATEST and
+                discover.LATEST in microversion[1:]):
+            raise TypeError(
+                "Specifying a '{major}.latest' microversion is not allowed.")
+        microversion = discover.version_to_string(microversion)
+        if not service_type:
+            if endpoint_filter and 'service_type' in endpoint_filter:
+                service_type = endpoint_filter['service_type']
+            else:
+                raise TypeError(
+                    "microversion {microversion} was requested but no"
+                    " service_type information is available. Either provide a"
+                    " service_type in endpoint_filter or pass"
+                    " microversion_service_type as an argument.".format(
+                        microversion=microversion))
+
+        # TODO(mordred) cinder uses volume in its microversion header. This
+        # logic should be handled in the future by os-service-types but for
+        # now hard-code for cinder.
+        if (service_type.startswith('volume') or
+                service_type == 'block-storage'):
+            service_type = 'volume'
+        headers.setdefault('OpenStack-API-Version',
+                           '{service_type} {microversion}'.format(
+                               service_type=service_type,
+                               microversion=microversion))
+        header_names = _mv_legacy_headers_for_service(service_type)
+        for h in header_names:
+            headers.setdefault(h, microversion)
+
     def request(self, url, method, json=None, original_ip=None,
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
-                endpoint_override=None, connect_retries=0, logger=_logger,
-                allow={}, **kwargs):
+                endpoint_override=None, connect_retries=0, logger=None,
+                allow=None, client_name=None, client_version=None,
+                microversion=None, microversion_service_type=None,
+                status_code_retries=0, retriable_status_codes=None,
+                **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -439,6 +629,24 @@ class Session(object):
         :type logger: logging.Logger
         :param dict allow: Extra filters to pass when discovering API
                            versions. (optional)
+        :param microversion: Microversion to send for this request.
+                       microversion can be given as a string or a tuple.
+                       (optional)
+        :param str microversion_service_type: The service_type to be sent in
+                       the microversion header, if a microversion is given.
+                       Defaults to the value of service_type from
+                       endpoint_filter if one exists. If endpoint_filter is not
+                       provided or does not have a service_type, microversion
+                       is given and microversion_service_type is not provided,
+                       an exception will be raised.
+        :param int status_code_retries: the maximum number of retries that
+                                        should be attempted for retriable
+                                        HTTP status codes (optional, defaults
+                                        to 0 - never retry).
+        :param list retriable_status_codes: list of HTTP status codes that
+                                            should be retried (optional,
+                                            defaults to HTTP 503, has no effect
+                                            when status_code_retries is 0).
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -453,7 +661,21 @@ class Session(object):
 
         :returns: The response to the request.
         """
+        # If a logger is passed in, use it and do not log requests, responses
+        # and bodies separately.
+        if logger:
+            split_loggers = False
+        else:
+            split_loggers = None
+        logger = logger or utils.get_logger(__name__)
+        # HTTP 503 - Service Unavailable
+        retriable_status_codes = retriable_status_codes or [503]
+
         headers = kwargs.setdefault('headers', dict())
+        if microversion:
+            self._set_microversion_headers(
+                headers, microversion, microversion_service_type,
+                endpoint_filter)
 
         if authenticated is None:
             authenticated = bool(auth or self.auth)
@@ -499,19 +721,55 @@ class Session(object):
         elif self.user_agent:
             user_agent = headers.setdefault('User-Agent', self.user_agent)
         else:
-            user_agent = headers.setdefault('User-Agent', DEFAULT_USER_AGENT)
+            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent should be
+            # ordered by decreasing significance.  If a user sets their product
+            # that value will be used. Otherwise we attempt to derive a useful
+            # product value. The value will be prepended it to the KSA version,
+            # requests version, and then the Python version.
+
+            agent = []
+
+            if self.app_name and self.app_version:
+                agent.append('%s/%s' % (self.app_name, self.app_version))
+            elif self.app_name:
+                agent.append(self.app_name)
+
+            if client_name and client_version:
+                agent.append('%s/%s' % (client_name, client_version))
+            elif client_name:
+                agent.append(client_name)
+
+            for additional in self.additional_user_agent:
+                agent.append('%s/%s' % additional)
+
+            if not agent:
+                # NOTE(jamielennox): determine_user_agent will return an empty
+                # string on failure so checking for None will ensure it is only
+                # called once even on failure.
+                if self._determined_user_agent is None:
+                    self._determined_user_agent = _determine_user_agent()
+
+                if self._determined_user_agent:
+                    agent.append(self._determined_user_agent)
+
+            agent.append(DEFAULT_USER_AGENT)
+            user_agent = headers.setdefault('User-Agent', ' '.join(agent))
 
         if self.original_ip:
             headers.setdefault('Forwarded',
                                'for=%s;by=%s' % (self.original_ip, user_agent))
 
         if json is not None:
-            headers['Content-Type'] = 'application/json'
+            headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
         for k, v in self.additional_headers.items():
             headers.setdefault(k, v)
 
+        # Bug #1766235: some headers may be bytes
+        headers = _sanitize_headers(headers)
+        kwargs['headers'] = headers
+
         kwargs.setdefault('verify', self.verify)
 
         if requests_auth:
@@ -528,7 +786,7 @@ class Session(object):
                                    data=kwargs.get('data'),
                                    headers=headers,
                                    query_params=query_params,
-                                   logger=logger)
+                                   logger=logger, split_loggers=split_loggers)
 
         # Force disable requests redirect handling. We will manage this below.
         kwargs['allow_redirects'] = False
@@ -538,7 +796,8 @@ class Session(object):
 
         send = functools.partial(self._send_request,
                                  url, method, redirect, log, logger,
-                                 connect_retries)
+                                 split_loggers, connect_retries,
+                                 status_code_retries, retriable_status_codes)
 
         try:
             connection_params = self.get_auth_connection_params(auth=auth)
@@ -554,6 +813,46 @@ class Session(object):
 
         resp = send(**kwargs)
 
+        # log callee and caller request-id for each api call
+        if log:
+            # service_name should be fetched from endpoint_filter if it is not
+            # present then use service_type as service_name.
+            service_name = None
+            if endpoint_filter:
+                service_name = endpoint_filter.get('service_name')
+                if not service_name:
+                    service_name = endpoint_filter.get('service_type')
+
+            # Nova uses 'x-compute-request-id' and other services like
+            # Glance, Cinder etc are using 'x-openstack-request-id' to store
+            # request-id in the header
+            request_id = (resp.headers.get('x-openstack-request-id') or
+                          resp.headers.get('x-compute-request-id'))
+            if request_id:
+                if self._get_split_loggers(split_loggers):
+                    id_logger = utils.get_logger(__name__ + '.request-id')
+                else:
+                    id_logger = logger
+                if service_name:
+                    id_logger.debug(
+                        '%(method)s call to %(service_name)s for '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'service_name': service_name,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
+                else:
+                    id_logger.debug(
+                        '%(method)s call to '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
+
         # handle getting a 401 Unauthorized response by invalidating the plugin
         # and then retrying the request. This is only tried once.
         if resp.status_code == 401 and authenticated and allow_reauth:
@@ -569,10 +868,25 @@ class Session(object):
                          resp.status_code)
             raise exceptions.from_response(resp, method, url)
 
+        if self._collect_timing:
+            for h in resp.history:
+                self._api_times.append(RequestTiming(
+                    method=h.request.method,
+                    url=h.request.url,
+                    elapsed=h.elapsed,
+                ))
+            self._api_times.append(RequestTiming(
+                method=resp.request.method,
+                url=resp.request.url,
+                elapsed=resp.elapsed,
+            ))
+
         return resp
 
-    def _send_request(self, url, method, redirect, log, logger,
-                      connect_retries, connect_retry_delay=0.5, **kwargs):
+    def _send_request(self, url, method, redirect, log, logger, split_loggers,
+                      connect_retries, status_code_retries,
+                      retriable_status_codes, connect_retry_delay=0.5,
+                      status_code_retry_delay=0.5, **kwargs):
         # NOTE(jamielennox): We handle redirection manually because the
         # requests lib follows some browser patterns where it will redirect
         # POSTs as GETs for certain statuses which is not want we want for an
@@ -617,13 +931,17 @@ class Session(object):
             time.sleep(connect_retry_delay)
 
             return self._send_request(
-                url, method, redirect, log, logger,
+                url, method, redirect, log, logger, split_loggers,
+                status_code_retries=status_code_retries,
+                retriable_status_codes=retriable_status_codes,
                 connect_retries=connect_retries - 1,
                 connect_retry_delay=connect_retry_delay * 2,
                 **kwargs)
 
         if log:
-            self._http_log_response(response=resp, logger=logger)
+            self._http_log_response(
+                response=resp, logger=logger,
+                split_loggers=split_loggers)
 
         if resp.status_code in self._REDIRECT_STATUSES:
             # be careful here in python True == 1 and False == 0
@@ -645,14 +963,34 @@ class Session(object):
                 # NOTE(jamielennox): We don't pass through connect_retry_delay.
                 # This request actually worked so we can reset the delay count.
                 new_resp = self._send_request(
-                    location, method, redirect, log, logger,
+                    location, method, redirect, log, logger, split_loggers,
                     connect_retries=connect_retries,
+                    status_code_retries=status_code_retries,
+                    retriable_status_codes=retriable_status_codes,
                     **kwargs)
 
                 if not isinstance(new_resp.history, list):
                     new_resp.history = list(new_resp.history)
                 new_resp.history.insert(0, resp)
                 resp = new_resp
+        elif (resp.status_code in retriable_status_codes and
+              status_code_retries > 0):
+
+            logger.info('Retriable status code %(code)s. Retrying in '
+                        '%(delay).1fs.',
+                        {'code': resp.status_code,
+                         'delay': status_code_retry_delay})
+            time.sleep(status_code_retry_delay)
+
+            # NOTE(jamielennox): We don't pass through connect_retry_delay.
+            # This request actually worked so we can reset the delay count.
+            return self._send_request(
+                url, method, redirect, log, logger, split_loggers,
+                connect_retries=connect_retries,
+                status_code_retries=status_code_retries - 1,
+                retriable_status_codes=retriable_status_codes,
+                status_code_retry_delay=status_code_retry_delay * 2,
+                **kwargs)
 
         return resp
 
@@ -767,9 +1105,70 @@ class Session(object):
         :returns: An endpoint if available or None.
         :rtype: string
         """
+        if 'endpoint_override' in kwargs:
+            return kwargs['endpoint_override']
+
         auth = self._auth_required(auth, 'determine endpoint URL')
+
         return auth.get_endpoint(self, **kwargs)
 
+    def get_endpoint_data(self, auth=None, **kwargs):
+        """Get endpoint data as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
+        :raises TypeError: If arguments are invalid
+
+        :returns: Endpoint data if available or None.
+        :rtype: keystoneauth1.discover.EndpointData
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_endpoint_data(self, **kwargs)
+
+    def get_api_major_version(self, auth=None, **kwargs):
+        """Get the major API version as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
+            plugin is not available.
+
+        :return: The major version of the API of the service discovered.
+        :rtype: tuple or None
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_api_major_version(self, **kwargs)
+
+    def get_all_version_data(self, auth=None, interface='public',
+                             region_name=None, **kwargs):
+        """Get version data for all services in the catalog.
+
+        :param auth:
+            The auth plugin to use for token. Overrides the plugin on
+            the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+        :param interface:
+            Type of endpoint to get version data for. Can be a single value
+            or a list of values. A value of None indicates that all interfaces
+            should be queried. (optional, defaults to public)
+        :param string region_name:
+            Region of endpoints to get version data for. A valueof None
+            indicates that all regions should be queried. (optional, defaults
+            to None)
+        :returns: A dictionary keyed by region_name with values containing
+            dictionaries keyed by interface with values being a list of
+            `~keystoneauth1.discover.VersionData`.
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_all_version_data(
+            self, interface=interface, region_name=region_name, **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -870,6 +1269,17 @@ class Session(object):
         auth = self._auth_required(auth, 'get project_id')
         return auth.get_project_id(self)
 
+    def get_timings(self):
+        """Return collected API timing information.
+
+        :returns: List of `RequestTiming` objects.
+        """
+        return self._api_times
+
+    def reset_timings(self):
+        """Clear API timing information."""
+        self._api_times = []
+
 
 REQUESTS_VERSION = tuple(int(v) for v in requests.__version__.split('.'))
 
@@ -879,7 +1289,7 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
 
     This Adapter also preserves the default behaviour of Requests which
     disables Nagle's Algorithm. See also:
-    http://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
+    https://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
     """
 
     def init_poolmanager(self, *args, **kwargs):
